/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

#pragma warning(disable:4786)
#pragma warning(disable:4503)

#include <stdarg.h>
#include <algorithm>
#include "interp.h"
#include "error.h"

/*
,   ,      .
        
   :    ,
 ,   . .   if 
while,      . ,     .
    ,   
    .
*/

/************************************************************************/
/*   CInterpreter                                       */
/************************************************************************/

CInterpreter::CInterpreter(Program* pProg) : 
	m_pProg(pProg)
{
	m_pStmt = NULL;
	FNMAP::iterator it = m_pProg->Functions.find("main");
	//      - - 'main'
	if (it == m_pProg->Functions.end())
	{
		Error(NULL, E_NO_MAIN);
	}
	else
	{
		//    
		InitIntProcs();
		//    
		PushSF(stFunction, it->second->Statements, it->second.get());
	}
}

//   
bool CInterpreter::DoStep()
{
	if (m_CodeStack.empty())
	{
		return false;
	}
	
	StackFrame& sf = m_CodeStack.top();
	if (sf.NextStmt == sf.End)
	{
		//      ,
		//    
		PopSF();
	}
	else
	{
		//   
		RunStatement((*sf.NextStmt).get());
	}

	return true;
}

//         
void CInterpreter::Error(const char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	char szBuf[2048] = {0};
	sprintf(
		szBuf, 
		m_pStmt ? "Runtime error in %s(%d): " : "Runtime error in %s: ", 
		m_pProg->sFileName.c_str(), 
		m_pStmt ? m_pStmt->nLine : 0);
	int nLen = strlen(szBuf);
	_vsnprintf(szBuf + nLen, sizeof(szBuf) - 1 - nLen, fmt, args);
	::Error("%s", szBuf);
	va_end(args);
}

//  r-,   
PVarVal CInterpreter::PopRVal(eVarType Type)
{
	PVarVal v = m_DataStack.top();
	m_DataStack.pop();
	if (Type != vtVar && v.Type == vtVar)
	{
		//    - ,   
		const char* pszVarName = v.strVal;
		v = m_CodeStack.top().pFn->Vars[pszVarName];
		if (v.Type == vtEmpty)
		{
			//   ,
			//    
			Error(E_EMPTY_VAR, pszVarName);
		}
	}
	//     
	else if (Type != vtEmpty && v.Type != Type)
	{
		//   
		if (!UnTypecast(v, Type))
		{
			Error(E_INV_OPTYPE);
		}
	}
	return v;
}

//  l-,   
PVarVal CInterpreter::PopLVal()
{
	PVarVal v = m_DataStack.top();
	m_DataStack.pop();
	if (v.Type != vtVar)
	{
		//    -  ,  
		Error(E_LVALUE);
	}
	return v;
}

//     
void CInterpreter::PushSF(eStmtType Type, STMTLIST& statements, PFunction* pFn)
{
	if (m_CodeStack.size() >= MAX_CODESTACK_DEEP)
	{
		//   
		Error(E_CODESTACK_DEEP);
	}

	StackFrame sf;
	sf.Type = Type;
	sf.pFn = pFn ? pFn : m_CodeStack.top().pFn;
	sf.NextStmt = statements.begin();
	sf.End = statements.end();
	m_CodeStack.push(sf);
}

//          
void CInterpreter::PopSF()
{
	if (!m_CodeStack.empty())
	{
		//     
		if (m_CodeStack.top().Type == stFunction)
		{
			//    
			m_CodeStack.top().pFn->Vars.clear();
		}
		m_CodeStack.pop();
	}
}

//      
void CInterpreter::FindFrameType(eStmtType Type, const char* err)
{
	//           
	//  ,  
	while (
		!m_CodeStack.empty() && 
		m_CodeStack.top().Type != Type && 
		m_CodeStack.top().Type != stFunction)
	{
		PopSF();
	}
	if (m_CodeStack.empty() || m_CodeStack.top().Type != Type)
	{
		//      
		Error(err);
	}
	else
	{
		//    
		PopSF();
	}
}

//  
void CInterpreter::UnOp(PVarVal (*OpFn)(PVarVal& x), const char* name)
{
	PVarVal x = PopRVal();
	try
	{
		m_DataStack.push(OpFn(x));
	}
	catch (__int64)
	{
		Error(E_INV_TYPEOP, x.Type, name);
	}
}

//  
void CInterpreter::BinOp(PVarVal (*OpFn)(PVarVal& x, PVarVal& y), 
						 const char* name)
{
	PVarVal y = PopRVal();
	PVarVal x = PopRVal();

	//   
	if (!BinTypecast(x, y))
	{
		Error(E_INV_TYPEOP, y.Type, name);
	}
	
	try
	{
		m_DataStack.push(OpFn(x, y));
	}
	catch (int)
	{
		Error(E_INV_TYPEOP, x.Type, name);
	}
}

//    
void CInterpreter::RunStatement(PStatement* stmt)
{
	//   ,  
	//        
	//    
	m_pStmt = stmt;
	//    
	switch (stmt->Type)
	{
		case stExpr:
			RunExprStmt((PExprStmt*)stmt);
			break;

		case stIf:
			RunIfStmt((PIfStmt*)stmt);
			break;

		case stWhile:
			RunWhileStmt((PWhileStmt*)stmt);
			break;

		case stBreak:
			RunBreakStmt();
			break;

		case stContinue:
			RunContinueStmt();
			break;

		case stReturn:
			RunReturnStmt((PReturnStmt*)stmt);
			break;

		default:
			Error(E_INV_STMT, stmt->Type);
	}
}

//  if
void CInterpreter::RunIfStmt(PIfStmt* stmt)
{
	NextStmt();
	if (ExprVal(stmt->Condition, vtBool).bVal)
	{
		//  
		PushSF(stIf, stmt->Statements);
	}
}

//  while
void CInterpreter::RunWhileStmt(PWhileStmt* stmt)
{
	if (ExprVal(stmt->Condition, vtBool).bVal)
	{
		//    
		PushSF(stWhile, stmt->Statements);
	}
	else
	{
		NextStmt();
	}
}

//  break
void CInterpreter::RunBreakStmt()
{
	FindFrameType(stWhile, E_INV_BREAK);
	NextStmt();
}

//  continue
void CInterpreter::RunContinueStmt()
{
	FindFrameType(stWhile, E_INV_CONTINUE);
}

//  return
void CInterpreter::RunReturnStmt(PReturnStmt* stmt)
{
	if (stmt->RetVal.Operators.empty())
	{
		//  return  
		m_DataStack.push(PVarVal());
	}
	else
	{
		//  return  
		RunExpr(stmt->RetVal);
		m_DataStack.push(PopRVal());
	}
	FindFrameType(stFunction, E_INV_RETURN);
}

//  ''
void CInterpreter::RunExprStmt(PExprStmt* stmt)
{
	NextStmt();
	RunExpr(stmt->Expr);
	PopRVal();
}

// 
void CInterpreter::RunExpr(PExpr& expr)
{
	OPLIST::iterator it = expr.Operators.begin();
	for (; it != expr.Operators.end(); ++it)
	{
		//        
		POperator& op = (*it);
		switch (op.Type)
		{
			case opAssign:
				{
					PVarVal v = PopRVal();
					PVarVal x = PopLVal();
					m_CodeStack.top().pFn->Vars[x.strVal] = v;
					m_DataStack.push(v);
				}
				break;

			case opAdd:
				BinOp(Add, "+");
				break;

			case opSub:
				BinOp(Sub, "-");
				break;

			case opMul:
				BinOp(Mul, "*");
				break;

			case opDiv:
				BinOp(Div, "/");
				break;

			case opNeg:
				UnOp(Neg, "unary -");
				break;

			case opAnd:
				BinOp(And, "&&");
				break;

			case opOr:
				BinOp(Or, "||");
				break;

			case opXor:
				BinOp(Xor, "^^");
				break;

			case opNot:
				UnOp(Not, "!");
				break;

			case opEq:
				BinOp(Eq, "==");
				break;

			case opNotEq:
				BinOp(NotEq, "!=");
				break;

			case opLess:
				BinOp(Less, "<");
				break;

			case opLessEq:
				BinOp(LessEq, "<=");
				break;

			case opGreater:
				BinOp(Greater, ">");
				break;

			case opGreaterEq:
				BinOp(GreaterEq, ">=");
				break;

			case opPushInt:
				PushC(vtInt, (__int64)op);
				break;

			case opPushFloat:
				PushC(vtFloat, (double)op);
				break;

			case opPushStr:
				PushC(vtString, (char*)op);
				break;

			case opPushBool:
				PushC(vtBool, (bool)op);
				break;

			case opPushVar:
				PushC(vtVar, (char*)op);
				break;

			case opCallSub:
				RunCallSub((char*)op);
				break;

			//   
			default:
				Error(E_INV_OP, op.Type);
		}
	}
}

//  
void CInterpreter::RunCallSub(const char* pszFnName)
{
	//    
	int nArgCount = PopRVal(vtInt);
	FNMAP::iterator it = m_pProg->Functions.find(pszFnName);

	//        
	if (it == m_pProg->Functions.end())
	{
		INTPROCSMAP::iterator it = m_IntProcs.find(pszFnName);
		//       
		if (it != m_IntProcs.end())
		{
			IntProc& intProc = it->second;
			
			//   
			if ((intProc.bVarArgs && nArgCount < intProc.nArgCount) ||
				(!intProc.bVarArgs && nArgCount != intProc.nArgCount))
			{
				Error(E_ARG_COUNT, pszFnName, nArgCount);
			}

			//  
			VARARGS args;
			for (int i = nArgCount - 1; i >= 0; i--)
			{
				args.push_back(intProc.bLValueArgs ? PopLVal() : PopRVal());
			}
			reverse(args.begin(), args.end());
			
			//   
			PVarVal v(((*this).*(intProc.pFn))(args));
			//     
			m_DataStack.push(v);
		}
		else
		{
			Error(E_NO_FUNC, pszFnName);
		}
	}
	else
	{
		PFunction* pFn = it->second.get();
		
		//   
		if (nArgCount < pFn->Args.size())
		{
			Error(E_ARG_COUNT, pszFnName, nArgCount);
		}

		//  
		for (int i = nArgCount - 1; i >= 0; i--)
		{
			pFn->Vars[pFn->Args[i]] = PopRVal();
		}
		
		//    
		PushSF(stFunction, pFn->Statements, pFn);

		//         
		int nSize = m_CodeStack.size();
		while (m_CodeStack.size() >= nSize)
		{
			StackFrame& sf = m_CodeStack.top();
			if (sf.NextStmt == sf.End)
			{
				PopSF();
			}
			else
			{
				RunStatement((*sf.NextStmt).get());
			}
		}
	}
}

//     
//   

PVarVal Add(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x + (__int64)y;
		case vtFloat:	return (double)x + (double)y;
		case vtString:	return (char*)((string((char*)x) + string((char*)y)).c_str());
	}
	throw 0;
}

PVarVal Sub(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x - (__int64)y;
		case vtFloat:	return (double)x - (double)y;
	}
	throw 0;
}

PVarVal Mul(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x * (__int64)y;
		case vtFloat:	return (double)x * (double)y;
	}
	throw 0;
}

PVarVal Div(PVarVal& x, PVarVal& y)
{
	try
	{
		switch (x.Type)
		{
			case vtInt:		return (__int64)x / (__int64)y;
			case vtFloat:	return (double)x / (double)y;
		}
	}
	catch (...)
	{
		Error(E_DIVIDE);
	}
	throw 0;
}

PVarVal Neg(PVarVal& x)
{
	switch (x.Type)
	{
		case vtInt:		return -(__int64)x;
		case vtFloat:	return -(double)x;
	}
	throw 0;
}

PVarVal And(PVarVal& x, PVarVal& y)
{
	if (!UnTypecast(x, vtBool))
	{
		throw 0;
	}
	return (bool)x && (bool)y;
}

PVarVal Or(PVarVal& x, PVarVal& y)
{
	if (!UnTypecast(x, vtBool))
	{
		throw 0;
	}
	return (bool)x || (bool)y;
}

PVarVal Xor(PVarVal& x, PVarVal& y)
{
	if (!UnTypecast(x, vtBool))
	{
		throw 0;
	}
	return bool((bool)x ^ (bool)y);
}

PVarVal Not(PVarVal& x)
{
	switch (x.Type)
	{
		case vtInt:		return !(__int64)x;
		case vtFloat:	return !(double)x;
		case vtString:	return !(char*)x[0];
		case vtBool:	return !(bool)x;
	}
	throw 0;
}

PVarVal Eq(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x == (__int64)y;
		case vtFloat:	return (double)x == (double)y;
		case vtString:	return strcmp((char*)x, (char*)y) == 0;
		case vtBool:	return (bool)x == (bool)y;
	}
	throw 0;
}

PVarVal NotEq(PVarVal& x, PVarVal& y)
{
	return !(bool)Eq(x, y);
}

PVarVal Less(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x < (__int64)y;
		case vtFloat:	return (double)x < (double)y;
		case vtString:	return strcmp((char*)x, (char*)y) < 0;
		case vtBool:	return (bool)x < (bool)y;
	}
	throw 0;
}

PVarVal LessEq(PVarVal& x, PVarVal& y)
{
	switch (x.Type)
	{
		case vtInt:		return (__int64)x <= (__int64)y;
		case vtFloat:	return (double)x <= (double)y;
		case vtString:	return strcmp((char*)x, (char*)y) <= 0;
		case vtBool:	return (bool)x <= (bool)y;
	}
	throw 0;
}

PVarVal Greater(PVarVal& x, PVarVal& y)
{
	return !(bool)LessEq(x, y);
}

PVarVal GreaterEq(PVarVal& x, PVarVal& y)
{
	return !(bool)Less(x, y);
}

bool UnTypecast(PVarVal& v, eVarType Type)
{
	if (v.Type != Type)
	{
		switch (Type)
		{
			case vtInt:
				switch (v.Type)
				{
					case vtFloat:	v = PVarVal((__int64)v.fltVal); break;
					case vtBool:	v = PVarVal((__int64)(v.bVal ? 1 : 0)); break;
					default:		return false;
				}
				break;

			case vtFloat:
				switch (v.Type)
				{
					case vtInt:		v = PVarVal((double)v.iVal); break;
					case vtBool:	v = PVarVal(v.bVal ? 1.0 : 0.0); break;
					default:		return false;
				}
				break;

			case vtBool:
				switch (v.Type)
				{
					case vtInt:		v = PVarVal(v.iVal ? true : false); break;
					case vtFloat:	v = PVarVal(v.fltVal ? true : false); break;
					case vtString:	v = PVarVal(v.strVal[0] ? true : false); break;
					default:		return false;
				}
				break;
		}
	}
	return true;
}

bool BinTypecast(PVarVal& x, PVarVal& y)
{
	if (y.Type != x.Type)
	{
		if (x.Type == vtFloat && y.Type == vtInt)
		{
			y.fltVal = y.iVal;
			y.Type = vtFloat;
		}
		else if (x.Type == vtInt && y.Type == vtFloat)
		{
			x.fltVal = x.iVal;
			x.Type = vtFloat;
		}
		else
		{
			return false;
		}
	}
	return true;
}
