/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

#pragma warning(disable:4786)

#include <fstream>
#include <stdarg.h>
#include "lexer.h"
#include "error.h"

#define RETTOKEN(X)		return CToken(X, m_nLine, m_nCol)

/************************************************************************/
/*   CLexer                                             */
/************************************************************************/

CLexer::CLexer(Program* pProg) :
	m_pProg(pProg),
	m_nLine(1),
	m_nCol(0),
	m_bUnget(false)
{
	m_pStream = new ifstream(pProg->sFileName.c_str(), ios::in);
	if (!*m_pStream)
	{
		::Error(E_OPEN_FILE, pProg->sFileName.c_str(), errno);
	}
}

CLexer::~CLexer()
{
	delete m_pStream;
}

//   
CToken CLexer::GetToken()
{
	char* p = NULL;
	eLexerState state = stNormal;
	char szBuf[MAX_STR_CONST + 1] = {0};
	int nIntVal = 0;
	double fltFloatVal = 0.0;
	bool bIsFloat = false;

	for (;;)
	{
		char c = get();

		//    -  ,  
		//       
		if (state >= stNumber && state <= stNumber6)
		{
			if (p - szBuf > MAX_STR_CONST)
			{
				Error(E_NUM_TOO_LONG);
			}
			else
			{
				*p++ = c;
			}
		}

		//       
		switch (state)
		{
			case stNormal:
				switch (c)
				{
					//  
					case ' ':
					case '\t':
					case '\v':
						break;

					//  
					case '\r':
						if (get() != '\n')
						{
							unget();
						}
					case '\n':
						m_nLine++, m_nCol = 0;
						break;

					// 
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
					case '.':
						unget();
						state = stNumber;
						p = szBuf;
						break;

					//  
					case '=':
						if (get() == '=')
						{
							RETTOKEN(ttOpEqual);
						}
						unget();
						RETTOKEN(ttOpAssign);

					case '+':
						RETTOKEN(ttOpPlus);

					case '-':
						RETTOKEN(ttOpMinus);

					case '*':
						RETTOKEN(ttOpMul);

					case '/':
						switch (get())
						{
							case '/':
								state = stComment1;
								break;

							case '*':
								state = stComment2;
								break;

							default:
								unget();
								RETTOKEN(ttOpDiv);
						}
						break;

					case '<':
						switch (get())
						{
							case '=': RETTOKEN(ttOpLessEq);
							case '>': RETTOKEN(ttOpNotEq);
						}
						unget();
						RETTOKEN(ttOpLess);

					case '>':
						if (get() == '=')
						{
							RETTOKEN(ttOpGreaterEq);
						}
						unget();
						RETTOKEN(ttOpGreater);

					case '&':
						if (get() != '&')
						{
							Error(E_NO_AND);
						}
						RETTOKEN(ttOpAnd);

					case '|':
						if (get() != '|')
						{
							Error(E_NO_OR);
						}
						RETTOKEN(ttOpOr);

					case '^':
						if (get() != '^')
						{
							Error(E_NO_XOR);
						}
						RETTOKEN(ttOpXor);
					
					case '!':
						if (get() != '=')
						{
							RETTOKEN(ttOpNot);
						}
						RETTOKEN(ttOpNotEq);

					// 
					case '(':
						RETTOKEN(ttDlmLBracket);

					case ')':
						RETTOKEN(ttDlmRBracket);

					case '{':
						RETTOKEN(ttDlmBegBlock);

					case '}':
						RETTOKEN(ttDlmEndBlock);

					case ',':
						RETTOKEN(ttDlmComma);

					case ';':
						RETTOKEN(ttDlmSemicolon);

					case '"':
						p = szBuf;
						state = stString;
						break;

					// EOF
					case EOF:
						RETTOKEN(ttEOF);

					default:
						// 
						if (isalpha(c) || c == '_')
						{
							unget();
							p = szBuf;
							state = stIdent;
							break;
						}
						//   -    
						Error(E_INVALID_SYM, (unsigned)c);
				}
				break;

			//  '//'
			case stComment1:
				if (c == '\r' || c == '\n')
				{
					unget();
					state = stNormal;
				}
				break;
				
			//  '/* */'
			case stComment2:
				switch (c)
				{
					case '*':
						if (get() == '/')
						{
							state = stNormal;
						}
						else
						{
							unget();
						}
						break;

					// new line
					case '\r':
						if (get() != '\n')
						{
							unget();
						}
					case '\n':
						m_nLine++, m_nCol = 0;
						break;
				}
				break;

			// 
			case stNumber:
				if (!isdigit(c))
				{
					switch (c)
					{
						case '.': 
							bIsFloat = true;
							state = stNumber2; 
							break;

						case 'e':
						case 'E':
							bIsFloat = true;
							state = stNumber4;
							break;
							
						default:
							return NumberToken(szBuf, false);
					}		
				}
				break;

			case stNumber2:
				if (isdigit(c))
				{
					state = stNumber3;
				}
				else
				{
					Error(E_INVALID_NUM, szBuf);
				}
				break;

			case stNumber3:
				if (c == 'e' || c == 'E')
				{
					state = stNumber4;
				}
				else if (!isdigit(c))
				{
					return NumberToken(szBuf, true);
				}
				break;

			case stNumber4:
				if (isdigit(c))
				{
					state = stNumber6;
				}
				else if (c == '+' || c == '-')
				{
					state = stNumber5;
				}
				else
				{
					Error(E_INVALID_NUM, szBuf);
				}
				break;

			case stNumber5:
				if (isdigit(c))
				{
					state = stNumber6;
				}
				else
				{
					Error(E_INVALID_NUM, szBuf);
				}
				break;

			case stNumber6:
				if (!isdigit(c))
				{
					return NumberToken(szBuf, true);
				}
				break;

			// 
			case stString:
				switch (c)
				{
					//   - 
					case '\r':
					case '\n':
						Error(E_STR_NEWLINE);

					//   - 
					case EOF:
						Error(E_EOF);

					//  ESC-
					case '\\':
						if (p - szBuf >= MAX_STR_CONST)
						{
							Error(E_STR_TOO_LONG);
						}
						switch (get())
						{				
							case '\'':	*p++ = '\'';	break;
							case '"':	*p++ = '"';		break;
							case 'a':	*p++ = '\a';	break;
							case 'b':	*p++ = '\b';	break;
							case 'f':	*p++ = '\f';	break;
							case 'n':	*p++ = '\n';	break;
							case 'r':	*p++ = '\r';	break;
							case 't':	*p++ = '\t';	break;
							case 'v':	*p++ = '\v';	break;
							
							case 'x':	
								state = stString2;
								break;

							default:
								Error(E_INVALID_ESC);
						}
						break;

					//  
					case '"':
						{
							CToken tk(ttStrConst, m_nLine, m_nCol);
							tk.strVal = m_pProg->StrTable.Add(szBuf);
							return tk;
						}

					//   
					default:
						if (p - szBuf >= MAX_STR_CONST)
						{
							Error(E_STR_TOO_LONG);
						}
						*p++ = c;
				}
				break;

			// ESC-
			case stString2:
				if (c == '\\')
				{
					if (p - szBuf >= MAX_STR_CONST)
					{
						Error(E_STR_TOO_LONG);
					}
					*p++ = '\\';
					state = stString;
				}
				else if ((c >= '0' && c <= '9') || 
					(c >= 'a' && c <= 'f') || 
					(c >= 'A' && c <= 'F'))
				{
					nIntVal = (unsigned)c;
					state = stString3;
				}
				else
				{
					Error(E_INVALID_ESC);
				}
				break;

			case stString3:
				if ((c >= '0' && c <= '9') || 
					(c >= 'a' && c <= 'f') || 
					(c >= 'A' && c <= 'F'))
				{
					if (p - szBuf >= MAX_STR_CONST)
					{
						Error(E_STR_TOO_LONG);
					}
					nIntVal |= (unsigned)c << 8;
					*p++ = strtoul((const char*)&nIntVal, NULL, 16);
					state = stString;
				}
				else
				{
					Error(E_INVALID_ESC);
				}
				break;

			// 
			case stIdent:
				if (isalpha(c) || c == '_')
				{
					if (p - szBuf >= MAX_IDENTIFIER)
					{
						Error(E_ID_TOO_LONG);
					}
					*p++ = c;
				}
				else
				{
					unget();
					
					// Keywords
					if (!strcmp(szBuf, "true"))
					{
						CToken tk(ttBoolConst, m_nLine, m_nCol);
						tk.bVal = true;
						return tk;
					}
					else if (!strcmp(szBuf, "false"))
					{
						CToken tk(ttBoolConst, m_nLine, m_nCol);
						tk.bVal = false;
						return tk;
					}
					if (!strcmp(szBuf, "function"))
						RETTOKEN(ttKwFunction);
					else if (!strcmp(szBuf, "if"))
						RETTOKEN(ttKwIf);
					else if (!strcmp(szBuf, "while"))
						RETTOKEN(ttKwWhile);
					else if (!strcmp(szBuf, "break"))
						RETTOKEN(ttKwBreak);
					else if (!strcmp(szBuf, "continue"))
						RETTOKEN(ttKwContinue);
					else if (!strcmp(szBuf, "return"))
						RETTOKEN(ttKwReturn);

					// Identifier
					CToken tk(ttIdent, m_nLine, m_nCol);
					tk.strVal = m_pProg->StrTable.Add(szBuf);
					return tk;
				}
		}
	}
}

//         
void CLexer::Error(const char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	::ErrorVarArg(m_pProg->sFileName.c_str(), m_nLine, m_nCol, fmt, args);
	va_end(args);
}

//     
CToken CLexer::NumberToken(char* szBuf, bool bIsFloat)
{
	unget();
	szBuf[strlen(szBuf) - 1] = '\0';

	CToken tk(bIsFloat ? ttFloatConst : ttIntConst, m_nLine, m_nCol);

	if (bIsFloat)
	{
		//    
		tk.fltVal = atof(szBuf);
	}
	else
	{
		//    
		sscanf(szBuf, "%I64i", &tk.iVal);
	}

	return tk;
}
