/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

#pragma warning(disable:4786)

#include "parser.h"
#include "error.h"
#include <algorithm>

/************************************************************************/
/*   CParser                                            */
/************************************************************************/

CParser::CParser(Program* pProg, CLexer* pLexer) :
	m_pProg(pProg), m_pLexer(pLexer)
{
	m_bUnget = false;
}

//         
void CParser::Error(const char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	::ErrorVarArg(
		m_pProg->sFileName.c_str(), 
		m_tkLast.nLine, 
		m_tkLast.nCol, 
		fmt, 
		args);
	va_end(args);
}

//       
CToken& CParser::Token(eTokenType type)
{
	if (get().Type != type)
	{
		Error(E_EXPECTED, TokenTypeToStr(type));
	}
	return m_tkLast;
}

//    
void CParser::Parse()
{
	FnList(m_pProg->Functions);
}

//  
void CParser::FnList(FNMAP& functions)
{
	while (get().Type != ttEOF)
	{
		unget();
		Token(ttKwFunction);
		Ident();

		auto_ptr<PFunction> f(new PFunction);
		const char* pszFnName = m_tkLast.strVal;

		//     
		if (functions.find(m_tkLast.strVal) != functions.end())
		{
			Error(E_DUP_FN, m_tkLast.strVal);
		}
		else
		{
			Token(ttDlmLBracket);

			//     
			if (get().Type != ttDlmRBracket)
			{
				unget();
				for (;;)
				{
					Ident();

					//      
					if (find(f->Args.begin(), f->Args.end(), m_tkLast.strVal) != f->Args.end())
					{
						Error(E_DUP_ARG, m_tkLast.strVal);
					}
					else
					{
						f->Args.push_back(m_tkLast.strVal);
					}
					
					if (get().Type == ttDlmRBracket)
					{
						break;
					}
					else
					{
						unget();
						Token(ttDlmComma);
					}
				}
			}

			//    
			StmtList(f->Statements);
			//   return    
			f->Statements.push_back(auto_ptr<PStatement>(new PReturnStmt));
		}

		//    
		functions[pszFnName] = f;
	}
}

//  
void CParser::StmtList(STMTLIST& statements)
{
	Token(ttDlmBegBlock);

	//    
	while (get().Type != ttDlmEndBlock)
	{
		auto_ptr<PStatement> pStmt;
		int nLine = m_tkLast.nLine;

		//       
		switch (m_tkLast.Type)
		{
			case ttKwIf:
				pStmt = auto_ptr<PStatement>(new PIfStmt);
				IfStmt(*(PIfStmt*)pStmt.get());
				break;

			case ttKwWhile:
				pStmt = auto_ptr<PStatement>(new PWhileStmt);
				WhileStmt(*(PWhileStmt*)pStmt.get());
				break;

			case ttKwBreak:
				pStmt = auto_ptr<PStatement>(new PBreakStmt);
				BreakStmt(*(PBreakStmt*)pStmt.get());
				break;

			case ttKwContinue:
				pStmt = auto_ptr<PStatement>(new PContinueStmt);
				ContinueStmt(*(PContinueStmt*)pStmt.get());
				break;

			case ttKwReturn:
				pStmt = auto_ptr<PStatement>(new PReturnStmt);
				ReturnStmt(*(PReturnStmt*)pStmt.get());
				break;

			//   
			default:
				unget();
				pStmt = auto_ptr<PStatement>(new PExprStmt);
				ExprStmt(*(PExprStmt*)pStmt.get());
				break;
		}

		//       - 
		//      
		pStmt->nLine = nLine;
		//    
		statements.push_back(pStmt);
	}
}

//  if
void CParser::IfStmt(PIfStmt& stmt)
{
	Token(ttDlmLBracket);
	Expr(stmt.Condition);
	Token(ttDlmRBracket);
	StmtList(stmt.Statements);
}

//  while
void CParser::WhileStmt(PWhileStmt& stmt)
{
	Token(ttDlmLBracket);
	Expr(stmt.Condition);
	Token(ttDlmRBracket);
	StmtList(stmt.Statements);
}

//  break
void CParser::BreakStmt(PBreakStmt& stmt)
{
	Token(ttDlmSemicolon);
}

//  continue
void CParser::ContinueStmt(PContinueStmt& stmt)
{
	Token(ttDlmSemicolon);
}

//  return
void CParser::ReturnStmt(PReturnStmt& stmt)
{
	get();
	if (m_tkLast.Type != ttDlmSemicolon)
	{
		unget();
		Expr(stmt.RetVal);
		Token(ttDlmSemicolon);
	}
}

//  ''
void CParser::ExprStmt(PExprStmt& stmt)
{
	Expr(stmt.Expr);
	Token(ttDlmSemicolon);
}

// 
void CParser::Expr(PExpr& expr)
{
	OpAssign(expr.Operators);
}

/*
    .               
(  ):                                         
          
     .   , ..      
   .              
                  
  :                 
  ,    -          
   .      
   .                              
*/

// 
void CParser::OpAssign(OPLIST& op)
{
	OpOr(op);
	while (get().Type == ttOpAssign)
	{
		OpAssign(op);
		op.push_back(opAssign);
	}
	unget();
}

//  ""
void CParser::OpOr(OPLIST& op)
{
	OpXor(op);
	while (get().Type == ttOpOr)
	{
		OpXor(op);
		op.push_back(opOr);
	}
	unget();
}

//  " "
void CParser::OpXor(OPLIST& op)
{
	OpAnd(op);
	while (get().Type == ttOpXor)
	{
		OpAnd(op);
		op.push_back(opXor);
	}
	unget();
}

//  ""
void CParser::OpAnd(OPLIST& op)
{
	OpEq(op);
	while (get().Type == ttOpAnd)
	{
		OpEq(op);
		op.push_back(opAnd);
	}
	unget();
}

//  -  
void CParser::OpEq(OPLIST& op)
{
	OpCmp(op);
	for (;;)
	{
		switch (get().Type)
		{
			case ttOpEqual:
				OpCmp(op);
				op.push_back(opEq);
				break;

			case ttOpNotEq:
				OpCmp(op);
				op.push_back(opNotEq);
				break;

			default:
				unget();
				return;
		}
	}
}

// ,   ..
void CParser::OpCmp(OPLIST& op)
{
	OpAdd(op);
	for (;;)
	{
		switch (get().Type)
		{
			case ttOpLess:
				OpAdd(op);
				op.push_back(opLess);
				break;

			case ttOpLessEq:
				OpAdd(op);
				op.push_back(opLessEq);
				break;

			case ttOpGreater:
				OpAdd(op);
				op.push_back(opGreater);
				break;

			case ttOpGreaterEq:
				OpAdd(op);
				op.push_back(opGreaterEq);
				break;

			default:
				unget();
				return;
		}
	}
}

//  - 
void CParser::OpAdd(OPLIST& op)
{
	OpMul(op);
	for (;;)
	{
		switch (get().Type)
		{
			case ttOpPlus:
				OpMul(op);
				op.push_back(opAdd);
				break;

			case ttOpMinus:
				OpMul(op);
				op.push_back(opSub);
				break;

			default:
				unget();
				return;
		}
	}
}

//  - 
void CParser::OpMul(OPLIST& op)
{
	OpNeg(op);
	for (;;)
	{
		switch (get().Type)
		{
			case ttOpMul:
				OpNeg(op);
				op.push_back(opMul);
				break;

			case ttOpDiv:
				OpNeg(op);
				op.push_back(opDiv);
				break;

			default:
				unget();
				return;
		}
	}
}

//  
void CParser::OpNeg(OPLIST& op)
{
	switch (get().Type)
	{
		case ttOpMinus:
			OpNot(op);
			op.push_back(opNeg);
			break;

		case ttOpPlus:
			OpNot(op);
			break;

		default:
			unget();
			OpNot(op);
	}
}

//  ""
void CParser::OpNot(OPLIST& op)
{
	if (get().Type == ttOpNot)
	{
		OpOperand(op);
		op.push_back(opNot);
	}
	else
	{
		unget();
		OpOperand(op);
	}
}

//  - ,      
void CParser::OpOperand(OPLIST& op)
{
	switch (get().Type)
	{
		case ttIdent:
			{
				const char* pszId = m_tkLast.strVal;
				if (get().Type != ttDlmLBracket)
				{
					//    
					unget();
					op.push_back(POperator(opPushVar, pszId));
				}
				else
				{
					//    
					OpCallSub(op, pszId);
				}
			}
			break;

		//   
		case ttDlmLBracket:
			OpAssign(op);
			Token(ttDlmRBracket);
			break;
			
		// 
		case ttIntConst:
			op.push_back(POperator(opPushInt, m_tkLast.iVal));
			break;
		case ttFloatConst:
			op.push_back(POperator(opPushFloat, m_tkLast.fltVal));
			break;
		case ttStrConst:
			op.push_back(POperator(opPushStr, m_tkLast.strVal));
			break;
		case ttBoolConst:
			op.push_back(POperator(opPushBool, m_tkLast.bVal));
			break;

		//  
		default:
			Error(E_NO_EXPR);
	}
}

//  
void CParser::OpCallSub(OPLIST& op, const char* pszName)
{
	int i = 0;
	//     
	if (get().Type != ttDlmRBracket)
	{
		unget();
		for (;;)
		{
			//    
			OpAssign(op);
			i++;
			if (get().Type == ttDlmRBracket)
			{
				break;
			}
			else
			{
				unget();
				Token(ttDlmComma);
			}
		}
	}

	//    
	op.push_back(POperator(opPushInt, i));
	//   
	op.push_back(POperator(opCallSub, pszName));
}
