/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

#pragma once
#pragma warning(disable:4786)

#include <map>
#include <list>
#include <vector>
#include <memory>
#include <string>
#include "strtable.h"

using namespace std;

/************************************************************************/
/*                            */
/************************************************************************/

//  
enum eStmtType
{
	stExpr,
	stIf,
	stWhile,
	stBreak,
	stContinue,
	stReturn,
	stFunction
};

//  
enum eOpType
{
	opAssign,
	opAdd,
	opSub,
	opMul,
	opDiv,
	opNeg,
	opAnd,
	opOr,
	opXor,
	opNot,
	opEq,
	opNotEq,
	opLess,
	opLessEq,
	opGreater,
	opGreaterEq,
	opPushInt,
	opPushFloat,
	opPushStr,
	opPushBool,
	opPushVar,
	opCallSub
};

//  
enum eVarType
{
	vtEmpty,
	vtInt,
	vtFloat,
	vtString,
	vtBool,
	vtVar
};

//  . 
//        ( VARIANT).
struct PVarVal
{
	eVarType	Type;		//  
	
	// 
	union
	{
		__int64	iVal;
		double	fltVal;
		char*	strVal;
		bool	bVal;
	};
	
	// 
	PVarVal(const PVarVal& v)
	{
		*this = v;
	}
	PVarVal(eVarType _Type = vtEmpty) : Type(_Type)
	{
		memset(&fltVal, 0, sizeof(fltVal));
	}
	~PVarVal()
	{
		if (Type == vtString || Type == vtVar)
		{
			free(strVal);
		}
	}
	
	//   
	PVarVal(__int64 v) :Type(vtInt), iVal(v) { }
	PVarVal(double v) :	Type(vtFloat), fltVal(v) { }
	PVarVal(char* v) :	Type(vtString), strVal(strdup(v)) { }
	PVarVal(bool v) :	Type(vtBool), bVal(v) { }

	operator =(const PVarVal& v)
	{
		Type = v.Type;
		if (v.Type == vtString || v.Type == vtVar)
		{
			strVal = strdup(v.strVal);
		}
		else
		{
			fltVal = v.fltVal;
		}
	}

	//   
	operator __int64()	const { return iVal; }
	operator double()	const { return fltVal; }
	operator char*()	const { return strVal; }
	operator bool()		const { return bVal; }
};

//     .
//    .
struct POperator
{
	eOpType	Type;	//  
	__int64	Arg;	// 
	POperator(eOpType _Type) : Type(_Type), Arg(0) { }
	template <class T>
	POperator(eOpType _Type, T _Arg) : Type(_Type), Arg(0)
	{
		memcpy(&Arg, &_Arg, sizeof(_Arg));
	}
	//   
	operator __int64()	const { return *(__int64*)&Arg; }
	operator double()	const { return *(double*)&Arg; }
	operator char*()	const { return *(char**)&Arg; }
	operator bool()		const { return *(bool*)&Arg; }
};

typedef vector<POperator> OPLIST;

// 
struct PExpr
{
	//  
	OPLIST Operators;
};

// 
struct PStatement
{
	int			nLine;
	eStmtType	Type;
	PStatement(eStmtType _Type) : Type(_Type) { }
};

typedef list<auto_ptr<PStatement> > STMTLIST;

// -
struct PExprStmt : public PStatement
{
	PExpr Expr;
	PExprStmt() : PStatement(stExpr) { }
};

//  if
struct PIfStmt : public PStatement
{
	PExpr		Condition;
	STMTLIST	Statements;
	PIfStmt() : PStatement(stIf) { }
};

//  while
struct PWhileStmt : public PStatement
{
	PExpr		Condition;
	STMTLIST	Statements;
	PWhileStmt() : PStatement(stWhile) { }
};

//  break
struct PBreakStmt : public PStatement
{
	PBreakStmt() : PStatement(stBreak) { }
};

//  continue
struct PContinueStmt : public PStatement
{
	PContinueStmt() : PStatement(stContinue) { }
};

//  return
struct PReturnStmt : public PStatement
{
	PExpr RetVal;
	PReturnStmt() : PStatement(stReturn) { }
};

typedef map<string,PVarVal>	VARSMAP;
typedef vector<string>		ARGSLIST;

// 
struct PFunction
{
	STMTLIST	Statements;		// 
	ARGSLIST	Args;			//   
	VARSMAP		Vars;			//    
};

typedef map<string,auto_ptr<PFunction> > FNMAP;

// 
struct Program
{
	string		sFileName;		//  
	CStrTable	StrTable;		//  
	FNMAP		Functions;		//  
};
